@@ -39,9 +39,7 @@ from __future__ import print_function, division, absolute_import
 import logging
 import socket
 import ssl
-import os
 import getpass
-import errno
 import uuid
 from struct import unpack
 from collections import deque
@@ -104,7 +102,7 @@ class _AddressList(object):
                            ' must be a host string or a (host, port) tuple')
                 self._logger.error(err_msg)
                 raise TypeError(err_msg)
-        
+
         self._logger.debug('Address list: {0}'.format(list(self.address_deque)))
 
     def _append(self, host, port):
@@ -197,7 +195,7 @@ class Connection(object):
         # Set up connection logger
         logger_name = 'vertica_{0}_{1}'.format(id(self), str(uuid.uuid4())) # must be a unique value
         self._logger = logging.getLogger(logger_name)
-        
+
         if 'log_level' not in self.options and 'log_path' not in self.options:
             # logger is disabled by default
             self._logger.disabled = True
@@ -215,6 +213,11 @@ class Connection(object):
         self._cursor = Cursor(self, self._logger, cursor_type=None,
                               unicode_error=self.options['unicode_error'])
 
+        # knob for using server-side prepared statements
+        self.options.setdefault('use_prepared_statements', False)
+        self._logger.debug('Connection prepared statements is {}'.format(
+                     'enabled' if self.options['use_prepared_statements'] else 'disabled'))
+
         self._logger.info('Connecting as user "{}" to database "{}" on host "{}" with port {}'.format(
                      self.options['user'], self.options['database'],
                      self.options['host'], self.options['port']))
@@ -325,6 +328,7 @@ class Connection(object):
 
     def balance_load(self, raw_socket):
         # Send load balance request and read server response
+        self._logger.debug('=> %s', messages.LoadBalanceRequest())
         raw_socket.sendall(messages.LoadBalanceRequest().get_message())
         response = raw_socket.recv(1)
 
@@ -335,6 +339,7 @@ class Connection(object):
                 self._logger.error(err_msg)
                 raise errors.MessageError(err_msg)
             res = BackendMessage.from_type(type_=response, data=raw_socket.recv(size-4))
+            self._logger.debug('<= %s', res)
             host = res.get_host()
             port = res.get_port()
             self._logger.info('Load balancing to host "{0}" on port {1}'.format(host, port))
@@ -350,6 +355,7 @@ class Connection(object):
             raw_socket.close()
             raw_socket = self.establish_connection()
         else:
+            self._logger.debug('<= LoadBalanceResponse: %s', response)
             self._logger.warning("Load balancing requested but not supported by server")
 
         return raw_socket
@@ -357,8 +363,10 @@ class Connection(object):
     def enable_ssl(self, raw_socket, ssl_options):
         from ssl import CertificateError, SSLError
         # Send SSL request and read server response
+        self._logger.debug('=> %s', messages.SslRequest())
         raw_socket.sendall(messages.SslRequest().get_message())
         response = raw_socket.recv(1)
+        self._logger.debug('<= SslResponse: %s', response)
         if response in ('S', b'S'):
             self._logger.info('Enabling SSL')
             try:
@@ -422,8 +430,8 @@ class Connection(object):
         if not isinstance(message, FrontendMessage):
             raise TypeError("invalid message: ({0})".format(message))
 
-        self._logger.debug('=> %s', message)
         sock = self._socket()
+        self._logger.debug('=> %s', message)
         try:
             for data in message.fetch_message():
                 try:
@@ -434,12 +442,8 @@ class Connection(object):
 
         except Exception as e:
             self.close_socket()
-            if str(e) == 'unsupported authentication method: 9':
-                raise errors.ConnectionError(
-                    'Error during authentication. Your password might be expired.')
-            else:
-                # noinspection PyTypeChecker
-                raise_from(errors.ConnectionError, e)
+            self._logger.error(str(e))
+            raise
 
     def close_socket(self):
         try:
@@ -488,6 +492,26 @@ class Connection(object):
                 break
         return message
 
+    def read_expected_message(self, expected_types, error_handler=None):
+        # Reads a message and does some basic error handling.
+        # expected_types must be a class (e.g. messages.BindComplete) or a tuple of classes
+        message = self.read_message()
+        if isinstance(message, expected_types):
+            return message
+        elif isinstance(message, messages.ErrorResponse):
+            if error_handler is not None:
+                error_handler(message)
+            else:
+                raise errors.DatabaseError(message.error_message())
+        else:
+            msg = 'Received unexpected message type: {}. '.format(type(message).__name__)
+            if isinstance(expected_types, tuple):
+                msg += 'Expected types: {}'.format(", ".join([t.__name__ for t in expected_types]))
+            else:
+                msg += 'Expected type: {}'.format(expected_types.__name__)
+            self._logger.error(msg)
+            raise errors.MessageError(msg)
+
     def process_message(self, message):
         if isinstance(message, messages.ErrorResponse):
             raise errors.ConnectionError(message.error_message())
@@ -544,10 +568,21 @@ class Connection(object):
 
             if isinstance(message, messages.Authentication):
                 # Password message isn't right format ("incomplete message from client")
-                if message.code != messages.Authentication.OK:
+                if message.code == messages.Authentication.OK:
+                    self._logger.info("User {} successfully authenticated"
+                        .format(self.options['user']))
+                elif message.code == messages.Authentication.CHANGE_PASSWORD:
+                    msg = "The password for user {} has expired".format(self.options['user'])
+                    self._logger.error(msg)
+                    raise errors.ConnectionError(msg)
+                elif message.code == messages.Authentication.PASSWORD_GRACE:
+                    self._logger.warning('The password for user {} will expire soon.'
+                        ' Please consider changing it.'.format(self.options['user']))
+                else:
                     self.write(messages.Password(password, message.code,
                                                  {'user': user,
-                                                  'salt': getattr(message, 'salt', None)}))
+                                                  'salt': getattr(message, 'salt', None),
+                                                  'usersalt': getattr(message, 'usersalt', None)}))
             else:
                 self.process_message(message)
 
