@@ -80,7 +80,7 @@ RE_BASIC_INSERT_STAT = (
     u"INSERT\\s+INTO\\s+(?P<target>({0}\\.)?{0})"
     u"\\s*\\(\\s*(?P<variables>{0}(\\s*,\\s*{0})*)\\s*\\)"
     u"\\s+VALUES\\s*\\(\\s*(?P<values>.*)\\s*\\)").format(RE_NAME)
-
+END_OF_RESULT_RESPONSES = (messages.CommandComplete, messages.PortalSuspended)
 
 class Cursor(object):
     # NOTE: this is used in executemany and is here for pandas compatibility
@@ -94,7 +94,8 @@ class Cursor(object):
         self._closed = False
         self._message = None
         self.operation = None
-
+        self.prepared_sql = None  # last statement been prepared
+        self.prepared_name = "s0"
         self.error = None
 
         #
@@ -121,6 +122,7 @@ class Cursor(object):
         raise errors.NotSupportedError('Cursor.callproc() is not implemented')
 
     def close(self):
+        self._close_prepared_statement()
         self._closed = True
 
     def cancel(self):
@@ -129,70 +131,85 @@ class Cursor(object):
 
         self.connection.close()
 
-    def execute(self, operation, parameters=None):
-        self.operation = as_text(operation)
+    def execute(self, operation, parameters=None, use_prepared_statements=None):
+        operation = as_text(operation)
+        self.operation = operation
 
         if self.closed():
             raise errors.Error('Cursor is closed')
 
         self.flush_to_query_ready()
 
-        if parameters:
-            # TODO: quote = True for backward compatibility. see if should be False.
-            operation = self.format_operation_with_parameters(operation, parameters)
-
         self.rowcount = -1
 
-        self.connection.write(messages.Query(operation))
+        use_prepared = bool(self.connection.options['use_prepared_statements']
+                if use_prepared_statements is None else use_prepared_statements)
+        if use_prepared:
+            # Execute the SQL as prepared statement (server-side bindings)
+            if parameters and not isinstance(parameters, (list, tuple)):
+                raise TypeError("Execute parameters should be a list/tuple")
 
-        # read messages until we hit an Error, DataRow or ReadyForQuery
-        self._message = self.connection.read_message()
-        while True:
-            if isinstance(self._message, messages.ErrorResponse):
-                raise errors.QueryError.from_error_response(self._message, operation)
-            elif isinstance(self._message, messages.RowDescription):
-                self.description = [Column(fd, self.unicode_error) for fd in self._message.fields]
-            elif isinstance(self._message, messages.DataRow):
-                break
-            elif isinstance(self._message, messages.ReadyForQuery):
-                break
-            elif isinstance(self._message, messages.CommandComplete):
-                break
-            else:
-                self.connection.process_message(self._message)
+            # If the SQL has not been prepared, prepare the SQL
+            if operation != self.prepared_sql:
+                self._prepare(operation)
+                self.prepared_sql = operation # the prepared statement is kept
 
-            self._message = self.connection.read_message()
+            # Bind the parameters and execute
+            self._execute_prepared_statement([parameters])
+        else:
+            # Execute the SQL directly (client-side bindings)
+            if parameters:
+                operation = self.format_operation_with_parameters(operation, parameters)
+            self._execute_simple_query(operation)
 
         return self
 
-    def executemany(self, operation, seq_of_parameters):
+    def executemany(self, operation, seq_of_parameters, use_prepared_statements=None):
         operation = as_text(operation)
+        self.operation = operation
 
         if not isinstance(seq_of_parameters, (list, tuple)):
             raise TypeError("seq_of_parameters should be list/tuple")
 
-        m = self._insert_statement.match(operation)
-        if m:
-            target = as_text(m.group('target'))
+        self.flush_to_query_ready()
+        use_prepared = bool(self.connection.options['use_prepared_statements']
+                if use_prepared_statements is None else use_prepared_statements)
 
-            variables = as_text(m.group('variables'))
-            variables = ",".join([variable.strip().strip('"') for variable in variables.split(",")])
-
-            values = as_text(m.group('values'))
-            values = ",".join([value.strip().strip('"') for value in values.split(",")])
-            seq_of_values = [self.format_operation_with_parameters(values, parameters, is_csv=True)
-                             for parameters in seq_of_parameters]
-            data = "\n".join(seq_of_values)
-
-            copy_statement = (
-                u"COPY {0} ({1}) FROM STDIN DELIMITER ',' ENCLOSED BY '\"' "
-                u"ENFORCELENGTH ABORT ON ERROR").format(target, variables)
-
-            self.copy(copy_statement, data)
+        if use_prepared:
+            # Execute the SQL as prepared statement (server-side bindings)
+            if len(seq_of_parameters) == 0:
+                raise ValueError("seq_of_parameters should not be empty")
+            if not all(isinstance(elem, (list, tuple)) for elem in seq_of_parameters):
+                raise TypeError("Each seq_of_parameters element should be a list/tuple")
+            # If the SQL has not been prepared, prepare the SQL
+            if operation != self.prepared_sql:
+                self._prepare(operation)
+                self.prepared_sql = operation # the prepared statement is kept
 
+            # Bind the parameters and execute
+            self._execute_prepared_statement(seq_of_parameters)
         else:
-            raise NotImplementedError(
-                "executemany is implemented for simple INSERT statements only")
+            m = self._insert_statement.match(operation)
+            if m:
+                target = as_text(m.group('target'))
+
+                variables = as_text(m.group('variables'))
+                variables = ",".join([variable.strip().strip('"') for variable in variables.split(",")])
+
+                values = as_text(m.group('values'))
+                values = ",".join([value.strip().strip('"') for value in values.split(",")])
+                seq_of_values = [self.format_operation_with_parameters(values, parameters, is_csv=True)
+                                 for parameters in seq_of_parameters]
+                data = "\n".join(seq_of_values)
+
+                copy_statement = (
+                    u"COPY {0} ({1}) FROM STDIN DELIMITER ',' ENCLOSED BY '\"' "
+                    u"ENFORCELENGTH ABORT ON ERROR").format(target, variables)
+
+                self.copy(copy_statement, data)
+            else:
+                raise NotImplementedError(
+                    "executemany is implemented for simple INSERT statements only")
 
     def fetchone(self):
         while True:
@@ -205,12 +222,19 @@ class Cursor(object):
                 # fetch next message
                 self._message = self.connection.read_message()
                 return row
+            elif isinstance(self._message, messages.RowDescription):
+                self.description = [Column(fd, self.unicode_error) for fd in self._message.fields]
             elif isinstance(self._message, messages.ReadyForQuery):
                 return None
-            elif isinstance(self._message, messages.CommandComplete):
+            elif isinstance(self._message, END_OF_RESULT_RESPONSES):
                 return None
+            elif isinstance(self._message, messages.EmptyQueryResponse):
+                pass
+            elif isinstance(self._message, messages.ErrorResponse):
+                raise errors.QueryError.from_error_response(self._message, self.operation)
             else:
-                self.connection.process_message(self._message)
+                raise errors.MessageError('Unexpected fetchone() state: {}'.format(
+                                    type(self._message).__name__))
 
             self._message = self.connection.read_message()
 
@@ -237,30 +261,44 @@ class Cursor(object):
         return list(self.iterate())
 
     def nextset(self):
+        """
+        Skip to the next available result set, discarding any remaining rows
+        from the current result set.
+
+        If there are no more result sets, this method returns False. Otherwise,
+        it returns a True and subsequent calls to the fetch*() methods will
+        return rows from the next result set.
+        """
         # skip any data for this set if exists
-        self.flush_to_command_complete()
+        self.flush_to_end_of_result()
 
         if self._message is None:
             return False
-        elif isinstance(self._message, messages.CommandComplete):
+        elif isinstance(self._message, END_OF_RESULT_RESPONSES):
             # there might be another set, read next message to find out
             self._message = self.connection.read_message()
             if isinstance(self._message, messages.RowDescription):
-                # next row will be either a DataRow or CommandComplete
+                self.description = [Column(fd, self.unicode_error) for fd in self._message.fields]
+                self._message = self.connection.read_message()
+                return True
+            elif isinstance(self._message, messages.BindComplete):
                 self._message = self.connection.read_message()
                 return True
             elif isinstance(self._message, messages.ReadyForQuery):
                 return False
+            elif isinstance(self._message, END_OF_RESULT_RESPONSES):
+                # result of a DDL/transaction
+                return True
             elif isinstance(self._message, messages.ErrorResponse):
                 raise errors.QueryError.from_error_response(self._message, self.operation)
             else:
-                raise errors.Error(
-                    'Unexpected nextset() state after CommandComplete: {0}'.format(self._message))
+                raise errors.MessageError(
+                    'Unexpected nextset() state after END_OF_RESULT_RESPONSES: {0}'.format(self._message))
         elif isinstance(self._message, messages.ReadyForQuery):
             # no more sets left to be read
             return False
         else:
-            raise errors.Error('Unexpected nextset() state: {0}'.format(self._message))
+            raise errors.MessageError('Unexpected nextset() state: {0}'.format(self._message))
 
     def setinputsizes(self, sizes):
         pass
@@ -284,15 +322,17 @@ class Cursor(object):
                 self._message = message
                 break
 
-    def flush_to_command_complete(self):
-        # if the last message isn't empty or CommandComplete, read messages until it is
-        if self._message is None or isinstance(self._message, (messages.ReadyForQuery,
-                                                               messages.CommandComplete)):
+    def flush_to_end_of_result(self):
+        # if the last message isn't empty or END_OF_RESULT_RESPONSES,
+        # read messages until it is
+        if (self._message is None or
+            isinstance(self._message, messages.ReadyForQuery) or
+            isinstance(self._message, END_OF_RESULT_RESPONSES)):
             return
 
         while True:
             message = self.connection.read_message()
-            if isinstance(message, messages.CommandComplete):
+            if isinstance(message, END_OF_RESULT_RESPONSES):
                 self._message = message
                 break
 
@@ -409,7 +449,7 @@ class Cursor(object):
 
             operation = operation % tuple(tlist)
         else:
-            raise errors.Error("Argument 'parameters' must be dict or tuple")
+            raise TypeError("Argument 'parameters' must be dict or tuple/list")
 
         return operation
 
@@ -419,3 +459,124 @@ class Cursor(object):
             return u'"{0}"'.format(re.escape(param))
         else:
             return QuotedString(param.encode(UTF_8, self.unicode_error)).getquoted()
+
+    def _execute_simple_query(self, query):
+        """
+        Send the query to the server using the simple query protocol.
+        Return True if this query contained no SQL (e.g. the string "--comment")
+        """
+        self._logger.info(u'Execute simple query: [{}]'.format(query))
+
+        # All of the statements in the query are sent here in a single message
+        self.connection.write(messages.Query(query))
+
+        # The first response could be a number of things:
+        #   ErrorResponse: Something went wrong on the server.
+        #   EmptyQueryResponse: The query being executed is empty.
+        #   RowDescription: This is the "normal" case when executing a query.
+        #                   It marks the start of the results.
+        #   CommandComplete: This occurs when executing DDL/transactions.
+        self._message = self.connection.read_message()
+        if isinstance(self._message, messages.ErrorResponse):
+            raise errors.QueryError.from_error_response(self._message, query)
+        elif isinstance(self._message, messages.RowDescription):
+            self.description = [Column(fd, self.unicode_error) for fd in self._message.fields]
+            self._message = self.connection.read_message()
+
+    def _error_handler(self, msg):
+        self.connection.write(messages.Sync())
+        raise errors.QueryError.from_error_response(msg, self.operation)
+
+    def _prepare(self, query):
+        """
+        Send the query to be prepared to the server. The server will parse the
+        query and return some metadata.
+        """
+        self._logger.info(u'Prepare a statement: [{}]'.format(query))
+
+        # Send Parse message to server
+        # We don't need to tell the server the parameter types yet
+        self.connection.write(messages.Parse(self.prepared_name, query, param_types=()))
+        # Send Describe message to server
+        self.connection.write(messages.Describe('prepared_statement', self.prepared_name))
+        self.connection.write(messages.Flush())
+
+        # Read expected message: ParseComplete
+        self._message = self.connection.read_expected_message(messages.ParseComplete, self._error_handler)
+
+        # Read expected message: ParameterDescription
+        self._message = self.connection.read_expected_message(messages.ParameterDescription, self._error_handler)
+        self._param_metadata = self._message.parameters
+
+        # Read expected message: RowDescription or NoData
+        self._message = self.connection.read_expected_message(
+                        (messages.RowDescription, messages.NoData), self._error_handler)
+        if isinstance(self._message, messages.NoData):
+            self.description = None # response was NoData for a DDL/transaction PreparedStatement
+        else:
+            self.description = [Column(fd, self.unicode_error) for fd in self._message.fields]
+
+        # Read expected message: CommandDescription
+        self._message = self.connection.read_expected_message(messages.CommandDescription, self._error_handler)
+        if len(self._message.command_tag) == 0:
+            msg = 'The statement being prepared is empty'
+            self._logger.error(msg)
+            self.connection.write(messages.Sync())
+            raise errors.EmptyQueryError(msg)
+
+        self._logger.info('Finish preparing the statement')
+
+    def _execute_prepared_statement(self, list_of_parameter_values):
+        """
+        Send multiple statement parameter sets to the server using the extended
+        query protocol. The server would bind and execute each set of parameter
+        values.
+
+        This function should not be called without first calling _prepare() to
+        prepare a statement.
+        """
+        portal_name = ""
+        parameter_type_oids = [metadata['data_type_oid'] for metadata in self._param_metadata]
+        parameter_count = len(self._param_metadata)
+
+        try:
+            if len(list_of_parameter_values) == 0:
+                raise ValueError("Empty list/tuple, nothing to execute")
+            for parameter_values in list_of_parameter_values:
+                if parameter_values is None:
+                    parameter_values = ()
+                self._logger.info(u'Bind parameters: {}'.format(parameter_values))
+                if len(parameter_values) != parameter_count:
+                    msg = ("Invalid number of parameters for {}: {} given, {} expected"
+                           .format(parameter_values, len(parameter_values), parameter_count))
+                    raise ValueError(msg)
+                self.connection.write(messages.Bind(portal_name, self.prepared_name,
+                                             parameter_values, parameter_type_oids))
+                self.connection.write(messages.Execute(portal_name, 0))
+            self.connection.write(messages.Sync())
+        except Exception as e:
+            self._logger.error(str(e))
+            # the server will not send anything until we issue a sync
+            self.connection.write(messages.Sync())
+            self._message = self.connection.read_message()
+            raise
+
+        self.connection.write(messages.Flush())
+
+        # Read expected message: BindComplete
+        self.connection.read_expected_message(messages.BindComplete)
+
+        self._message = self.connection.read_message()
+        if isinstance(self._message, messages.ErrorResponse):
+            raise errors.QueryError.from_error_response(self._message, self.prepared_sql)
+
+    def _close_prepared_statement(self):
+        """
+        Close the prepared statement on the server.
+        """
+        self.prepared_sql = None
+        self.flush_to_query_ready()
+        self.connection.write(messages.Close('prepared_statement', self.prepared_name))
+        self.connection.write(messages.Flush())
+        self._message = self.connection.read_expected_message(messages.CloseComplete)
+        self.connection.write(messages.Sync())
